class ResourceController:
    def __init__(self, duct):
        self.duct = duct

    async def get_event_history(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["EVENT_HISTORY"], None)

    async def set_event_history(self, eid, query):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["EVENT_HISTORY"], [eid, query])

    async def list_projects(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["LIST_PROJECTS"], None)

    async def create_project(self, ProjectName):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["CREATE_PROJECT"], { "ProjectName": ProjectName })

    async def get_project_scheme(self, ProjectName, Cached):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["GET_PROJECT_SCHEME"],
                                        { "ProjectName": ProjectName, "Cached": Cached })

    async def create_templates(self, ProjectName, TemplateNames, PresetEnvName, PresetTemplateName):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["CREATE_TEMPLATES"],
                                        { "ProjectName": ProjectName, "TemplateNames": TemplateNames,
                                          "PresetEnvName": PresetEnvName, "PresetTemplateName": PresetTemplateName })

    async def list_template_presets(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["LIST_TEMPLATE_PRESETS"], None)

    async def list_templates(self, ProjectName):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["LIST_TEMPLATES"], { "ProjectName": ProjectName })

    async def get_responses_for_template(self, ProjectName, TemplateName, PageStart, PageLength):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["GET_RESPONSES_FOR_TEMPLATE"],
                                        { "ProjectName": ProjectName, "TemplateName": TemplateName, "PageStart": PageStart, "PageLength": PageLength })

    async def get_responses_for_nanotask(self, NanotaskId):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["GET_RESPONSES_FOR_NANOTASK"], { "NanotaskId": NanotaskId })

    async def get_nanotasks(self, ProjectName, TemplateName):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["GET_NANOTASKS"],
                                        { "ProjectName": ProjectName, "TemplateName": TemplateName })

    async def upload_nanotasks(self, ProjectName, TemplateName, Nanotasks, NumAssignable, Priority, TagName):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["UPLOAD_NANOTASKS"],
                                        { "ProjectName": ProjectName, "TemplateName": TemplateName,
                                          "Nanotasks": Nanotasks, "NumAssignable": NumAssignable,
                                          "Priority": Priority, "TagName": TagName })

    async def delete_nanotasks(self, ProjectName, TemplateName, NanotaskIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["DELETE_NANOTASKS"],
                                        { "ProjectName": ProjectName, "TemplateName": TemplateName, "NanotaskIds": NanotaskIds })

    async def update_nanotask_num_assignable(self, ProjectName, TemplateName, NanotaskId, NumAssignable):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["UPDATE_NANOTASK_NUM_ASSIGNABLE"],
                                        { "ProjectName": ProjectName, "TemplateName": TemplateName,
                                          "NanotaskId": NanotaskId, "NumAssignable": NumAssignable })

    async def create_session(self, ProjectName, PlatformWorkerId, ClientToken, Platform):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["SESSION"], { "Command": "Create",
                                        "ProjectName": ProjectName, "TemplateName": TemplateName,
                                        "NanotaskId": NanotaskId, "NumAssignable": NumAssignable })

    async def get_template_node(self, Target, WorkSessionId, NodeSessionId):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["SESSION"], { "Command": "Get",
                                        "Target": Target, "WorkSessionId": WorkSessionId, "NodeSessionId": NodeSessionId })

    async def set_response(self, WorkSessionId, NodeSessionId, Answers):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["SESSION"], { "Command": "SetResponse",
                                        "WorkSessionId": WorkSessionId, "NodeSessionId": NodeSessionId, "Answers": Answers })

    async def check_platform_worker_id_existence_for_project(self, ProjectName, Platform, PlatformWorkerId):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["CHECK_PLATFORM_WORKER_ID_EXISTENCE_FOR_PROJECT"],
                                        { "ProjectName": ProjectName, "Platform": Platform, "PlatformWorkerId": PlatformWorkerId })


class MTurkController:
    def __init__(self, duct):
        self.duct = duct

    async def get_credentials(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_GET_CREDENTIALS"], None)

    async def set_credentials(self, AccessKeyId, SecretAccessKey):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_SET_CREDENTIALS"],
                                        { "AccessKeyId": AccessKeyId, "SecretAccessKey": SecretAccessKey })

    async def clear_credentials(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_CLEAR_CREDENTIALS"], None)

    async def set_sandbox(self, Enabled):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_SET_SANDBOX"], { "Enabled": Enabled })

    async def get_hit_types(self, HITTypeIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_GET_HIT_TYPES"], { "HITTypeIds": HITTypeIds })

    async def create_hit_type(self, CreateHITTypeParams, HITTypeQualificationTypeId):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_CREATE_HIT_TYPE"],
                                        { "CreateHITTypeParams": CreateHITTypeParams,
                                          "HITTypeQualificationTypeId": HITTypeQualificationTypeId })

    async def create_hits_with_hit_type(self, ProjectName, NumHITs, CreateHITsWithHITTypeParams):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_CREATE_HITS_WITH_HIT_TYPE"], 
                                        { "ProjectName": ProjectName, "NumHITs": NumHITs,
                                          "CreateHITsWithHITTypeParams": CreateHITsWithHITTypeParams })

    async def list_qualifications(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_LIST_QUALIFICATIONS"], None)

    async def list_hits(self, Cached):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_LIST_HITS"], { "Cached": Cached })

    async def list_hits_for_hit_type(self, HITTypeId=None, Cached=True):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_LIST_HITS_FOR_HIT_TYPE"],
                                        { "HITTypeId": HITTypeId, "Cached": Cached })

    async def expire_hits(self, HITIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_EXPIRE_HITS"], { "HITIds": HITIds })

    async def delete_hits(self, HITIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_DELETE_HITS"], { "HITIds": HITIds })

    async def create_qualification(self, QualificationTypeParams):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_CREATE_QUALIFICATION"],
                                        { "QualificationTypeParams": QualificationTypeParams })

    async def associate_qualifications_with_workers(self, QualificationTypeId, WorkerIds, IntegerValue, SendNotification):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_ASSOCIATE_QUALIFICATIONS_WITH_WORKERS"],
                                        { "QualificationTypeId": QualificationTypeId, "WorkerIds": WorkerIds, "IntegerValue": IntegerValue, "SendNotification": SendNotification })

    async def list_workers(self):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["LIST_WORKERS"], { "Platform": "MTurk" })

    async def list_workers_with_qualification_type(self, QualificationTypeId):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_LIST_WORKERS_WITH_QUALIFICATION_TYPE"],
                                        { "QualificationTypeId": QualificationTypeId })

    async def delete_qualifications(self, QualificationTypeIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_DELETE_QUALIFICATIONS"],
                                        { "QualificationTypeIds": QualificationTypeIds })

    async def list_assignments(self, Cached):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_LIST_ASSIGNMENTS"], { "Cached": Cached })

    async def list_assignments_for_hits(self, HITIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_LIST_ASSIGNMENTS_FOR_HITS"], { "HITIds": HITIds })

    async def approve_assignments(self, AssignmentIds, RequesterFeedback):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_APPROVE_ASSIGNMENTS"],
                                        { "AssignmentIds": AssignmentIds, "RequesterFeedback": RequesterFeedback })

    async def reject_assignments(self, AssignmentIds, RequesterFeedback):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_REJECT_ASSIGNMENTS"],
                                        { "AssignmentIds": AssignmentIds, "RequesterFeedback": RequesterFeedback })

    async def get_assignments(self, AssignmentIds):
        await self.duct.send(self.duct.next_rid(), self.duct.EVENT["MTURK_GET_ASSIGNMENTS"], { "AssignmentIds": AssignmentIds })




import functools

from ducts_client import Duct
from .controllers import ResourceController, MTurkController
from .event_listeners import ResourceEventListener, MTurkEventListener

class TuttiDuct(Duct):
    def __init__(self):
        super().__init__()
        self.onopen_handlers = [];
        self.controllers = {
            "resource": ResourceController(self),
            "mturk": MTurkController(self)
        }
        self.event_listeners = {
            "resource": ResourceEventListener(),
            "mturk": MTurkEventListener()
        }

    def add_onopen_handler(self, handler_coro):
        self.onopen_handlers.append(handler_coro)

    async def _onopen(self, event):
        await super()._onopen(event)

        self.setup_handlers()
        for handler in self.onopen_handlers:
            await handler()

    def setup_handlers(self):
        self.set_event_handler( self.EVENT["EVENT_HISTORY"], self.handle_for_event_history )

        self.set_event_handler( self.EVENT["LIST_PROJECTS"],
                                    functools.partial(self.handle_for, "resource", "list_projects") )
        self.set_event_handler( self.EVENT["CREATE_PROJECT"],
                                    functools.partial(self.handle_for, "resource", "create_project") )
        self.set_event_handler( self.EVENT["GET_PROJECT_SCHEME"],
                                    functools.partial(self.handle_for, "resource", "get_project_scheme") )
        self.set_event_handler( self.EVENT["CREATE_TEMPLATES"],
                                    functools.partial(self.handle_for, "resource", "create_templates") )
        self.set_event_handler( self.EVENT["LIST_TEMPLATE_PRESETS"],
                                    functools.partial(self.handle_for, "resource", "list_template_presets") )
        self.set_event_handler( self.EVENT["LIST_TEMPLATES"],
                                    functools.partial(self.handle_for, "resource", "list_templates") )
        self.set_event_handler( self.EVENT["GET_RESPONSES_FOR_TEMPLATE"],
                                    functools.partial(self.handle_for, "resource", "get_responses_for_template") )
        self.set_event_handler( self.EVENT["GET_RESPONSES_FOR_NANOTASK"],
                                    functools.partial(self.handle_for, "resource", "get_responses_for_nanotask") )
        self.set_event_handler( self.EVENT["GET_NANOTASKS"],
                                    functools.partial(self.handle_for, "resource", "get_nanotasks") )
        self.set_event_handler( self.EVENT["UPLOAD_NANOTASKS"],
                                    functools.partial(self.handle_for, "resource", "upload_nanotasks") )
        self.set_event_handler( self.EVENT["DELETE_NANOTASKS"],
                                    functools.partial(self.handle_for, "resource", "delete_nanotasks") )
        self.set_event_handler( self.EVENT["UPDATE_NANOTASK_NUM_ASSIGNABLE"],
                                    functools.partial(self.handle_for, "resource", "update_nanotask_num_assignable") )
        self.set_event_handler( self.EVENT["SESSION"], self.handle_for_session )
        self.set_event_handler( self.EVENT["CHECK_PLATFORM_WORKER_ID_EXISTENCE_FOR_PROJECT"],
                                    functools.partial(self.handle_for, "resource", "check_platform_worker_id_existence_for_project") )

        self.set_event_handler( self.EVENT["MTURK_GET_CREDENTIALS"],
                                    functools.partial(self.handle_for, "mturk", "get_credentials") )
        self.set_event_handler( self.EVENT["MTURK_SET_CREDENTIALS"],
                                    functools.partial(self.handle_for, "mturk", "set_credentials") )
        self.set_event_handler( self.EVENT["MTURK_CLEAR_CREDENTIALS"],
                                    functools.partial(self.handle_for, "mturk", "clear_credentials") )
        self.set_event_handler( self.EVENT["MTURK_SET_SANDBOX"],
                                    functools.partial(self.handle_for, "mturk", "set_sandbox") )
        self.set_event_handler( self.EVENT["MTURK_GET_HIT_TYPES"],
                                    functools.partial(self.handle_for, "mturk", "get_hit_types") )
        self.set_event_handler( self.EVENT["MTURK_CREATE_HIT_TYPE"],
                                    functools.partial(self.handle_for, "mturk", "create_hit_type") )
        self.set_event_handler( self.EVENT["MTURK_CREATE_HITS_WITH_HIT_TYPE"],
                                    functools.partial(self.handle_for, "mturk", "create_hits_with_hit_type") )
        self.set_event_handler( self.EVENT["MTURK_LIST_QUALIFICATIONS"],
                                    functools.partial(self.handle_for, "mturk", "list_qualifications") )
        self.set_event_handler( self.EVENT["MTURK_LIST_HITS"],
                                    functools.partial(self.handle_for, "mturk", "list_hits") )
        self.set_event_handler( self.EVENT["MTURK_LIST_HITS_FOR_HIT_TYPE"],
                                    functools.partial(self.handle_for, "mturk", "list_hits_for_hit_type") )
        self.set_event_handler( self.EVENT["MTURK_EXPIRE_HITS"],
                                    functools.partial(self.handle_for, "mturk", "expire_hits") )
        self.set_event_handler( self.EVENT["MTURK_DELETE_HITS"],
                                    functools.partial(self.handle_for, "mturk", "delete_hits") )
        self.set_event_handler( self.EVENT["MTURK_CREATE_QUALIFICATION"],
                                    functools.partial(self.handle_for, "mturk", "create_qualification") )
        self.set_event_handler( self.EVENT["MTURK_ASSOCIATE_QUALIFICATIONS_WITH_WORKERS"],
                                    functools.partial(self.handle_for, "mturk", "associate_qualifications_with_workers") )
        self.set_event_handler( self.EVENT["LIST_WORKERS"],
                                    functools.partial(self.handle_for, "mturk", "list_workers") )
        self.set_event_handler( self.EVENT["MTURK_LIST_WORKERS_WITH_QUALIFICATION_TYPE"],
                                    functools.partial(self.handle_for, "mturk", "list_workers_with_qualification_type") )
        self.set_event_handler( self.EVENT["MTURK_DELETE_QUALIFICATIONS"],
                                    functools.partial(self.handle_for, "mturk", "delete_qualifications") )
        self.set_event_handler( self.EVENT["MTURK_LIST_ASSIGNMENTS"],
                                    functools.partial(self.handle_for, "mturk", "list_assignments") )
        self.set_event_handler( self.EVENT["MTURK_LIST_ASSIGNMENTS_FOR_HITS"],
                                    functools.partial(self.handle_for, "mturk", "list_assignments_for_hits") )
        self.set_event_handler( self.EVENT["MTURK_APPROVE_ASSIGNMENTS"],
                                    functools.partial(self.handle_for, "mturk", "approve_assignments") )
        self.set_event_handler( self.EVENT["MTURK_REJECT_ASSIGNMENTS"],
                                    functools.partial(self.handle_for, "mturk", "reject_assignments") )
        self.set_event_handler( self.EVENT["MTURK_GET_ASSIGNMENTS"],
                                    functools.partial(self.handle_for, "mturk", "get_assignments") )


    async def handle_for(self, _type, event_name, rid, eid, data):
        await self._handle(_type, event_name, data)

    async def handle_for_event_history(self, rid, eid, data):
        if "AllHistory" in data["Contents"]:  await self._handle("resource", "get_event_history", data)
        elif "History" in data["Contents"]:   await self._handle("resource", "set_event_history", data)

    async def handle_for_session(self, rid, eid, data):
        if data["Contents"]["Command"]=="Create":  await self._handle("resource", "create_session", data)
        elif data["Contents"]["Command"]=="Get":  await self._handle("resource", "get_template_node", data)
        elif data["Contents"]["Command"]=="SetResponse":  await self._handle("resource", "set_response", data)

    async def _handle(self, _type, name, data):
        if data["Status"]=="Success":
            for func in self.event_listeners[_type].get_handlers(name):  await func(data["Contents"], is_error=False)
        else:
            for func in self.event_listeners[_type].get_handlers(name):  await func(data, is_error=True)



import ducts_client.event_listeners

class DuctEventListener(ducts_client.event_listeners.DuctEventListener):
    def __init__(self):
        super().__init__()
        self.handlers = {}

    def add_handler(self, event_name, handler):
        self.handlers[event_name].append(handler)

    def get_handlers(self, event_name):
        if event_name in self.handlers:
            return self.handlers[event_name]
        else:
            raise Exception(f"Event named '{event_name}' is not found")
        
    def on(self, names, handler):
        if isinstance(names, str):  names = [names]
        for name in names:  self.add_handler(name, handler)

class ResourceEventListener(DuctEventListener):
    def __init__(self):
        super().__init__()

        self.handlers = {
            "list_projects": [],
            "create_project": [],
            "get_event_history": [],
            "set_event_history": [],
            "list_projects": [],
            "create_project": [],
            "get_project_scheme": [],
            "create_templates": [],
            "list_template_presets": [],
            "list_templates": [],
            "get_responses_for_template": [],
            "get_responses_for_nanotask": [],
            "get_nanotasks": [],
            "upload_nanotasks": [],
            "delete_nanotasks": [],
            "update_nanotask_num_assignable": [],
            "create_session": [],
            "get_template_node": [],
            "set_response": [],
            "check_platform_worker_id_existence_for_project": [],
        }

class MTurkEventListener(DuctEventListener):
    def __init__(self):
        super().__init__()

        self.handlers = {
            "get_credentials": [],
            "set_credentials": [],
            "clear_credentials": [],
            "set_sandbox": [],
            "get_hit_types": [],
            "create_hit_type": [],
            "create_hits_with_hit_type": [],
            "list_qualifications": [],
            "list_hits": [],
            "list_hits_for_hit_type": [],
            "expire_hits": [],
            "delete_hits": [],
            "create_qualification": [],
            "associate_qualifications_with_workers": [],
            "list_workers": [],
            "list_workers_with_qualification_type": [],
            "delete_qualifications": [],
            "list_assignments": [],
            "list_assignments_for_hits": [],
            "approve_assignments": [],
            "reject_assignments": [],
            "get_assignments": [],
        }
